/*
** Module   : AppServerDefinition.java
** Abstract : The settings of an appserver.
**
** Copyright (c) 2013, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- ---------------------------------Description----------------------------------
** 001 CA  20130704 Created initial version.
*/

package com.goldencode.p2j.util;

import java.util.*;
import org.apache.commons.logging.*;
import com.goldencode.p2j.directory.*;

/**
 * A container of all the settings of an appservers.  Access to the getters and setters is 
 * synchronized, so that in case the setting is changed/refreshed by another thread, concurrent 
 * access is safe.
 */
public class AppServerDefinition
{
   /** Logger. */
   private static final Log LOG = LogFactory.getLog(AppServerDefinition.class);

   /** A map of all initialized appserver definitions. */
   private static final Map<String, AppServerDefinition> definitions = 
      new HashMap<String, AppServerDefinition>();
   
   /** Flag indicating if this appserver is automatically started at the P2J server startup. */
   private boolean autoStart;
   
   /** The dir where this appserver will be started. */
   private String userDir;
   
   /**
    * The JVM arguments needed to start this appserver. This includes any environment variables,
    * which will be specified via <code>-Dprop=value</code> arguments.
    */
   private String[] jvmArguments;

   /** The appserver's name. */
   private String appServer;

   /** The appserver's operating mode. */
   private AppServerOperatingMode operatingMode;
   
   /**
    * The amount of time (in seconds) to wait for an agent to be available, before attempting
    * to auto-start a new one. 
    */
   private Integer requestTimeout;
   
   /**
    * The number of seconds waiting before auto-trimming the agents, if they exceed the
    * minimum capacity.
    */
   private Integer autoTrimTimeout;
   
   /** Appserver's explicit propath. */
   private String propath;
   
   /** The number of agents to start initially. */
   private Integer initialAgents;
   
   /** The minimum number of agents required for this appserver. */
   private Integer minAgents;
   
   /** The maximum number of agents which can be alive. */
   private Integer maxAgents;
   
   /** The activate procedure; may be null/empty. */
   private String activate;
   
   /** The deactivate procedure; may be null/empty. */
   private String deactivate;
   
   /** The connect procedure; may be null/empty. */
   private String connect;
   
   /** The disconnect procedure; may be null/empty. */
   private String disconnect;
   
   /** The startup procedure; may be null/empty. */
   private String startup;
   
   /** The shutdown procedure; may be null/empty. */
   private String shutdown;
   
   /** The parameter for the startup procedure; may be null/empty. */
   private String startupParameter;

   /**
    * Read the definition for the given appserver, and save it in this instance.
    * 
    * @param    appServer
    *           The application's server name.
    * 
    * @throws   IllegalArgumentException
    *           If no appserver (with the given name) is configured in the directory.
    */
   private AppServerDefinition(String appServer)
   {
      this.appServer = appServer;
      
      refresh();
   }

   /**
    * Get the settings for the specified appserver.  If not yet initialized, it will read the
    * settings for the appserver from the directory.  The created instance will be shared by all
    * parties.
    * <p>
    * The initial call (which creates the instance) must be coming from a thread with its context
    * authorized to access the <code>appservers</code> node for the P2J server.
    * 
    * @param    appServer
    *           The name of the appserver.
    *           
    * @return   The instance containing the settings for the specified appserver.
    */
   public static AppServerDefinition get(String appServer)
   {
      AppServerDefinition def = null;

      synchronized (definitions)
      {
         def = definitions.get(appServer);
         if (def == null)
         {
            def = new AppServerDefinition(appServer);
            definitions.put(appServer, def);
         }
      }
      
      return def;
   }

   /**
    * Remove the definition for the given appserver.
    * 
    * @param    appServer
    *           The appserver name.
    */
   static void remove(String appServer)
   {
      synchronized (definitions)
      {
         definitions.remove(appServer);
      }
   }
   
   /**
    * Refresh this appserver instance from the directory.  Must be called from threads which have
    * their context authorized to access the <code>appservers</code> node for the P2J server.
    */
   public synchronized void refresh()
   {
      // access the directory service
      DirectoryService ds = DirectoryService.getInstance();
      
      if (!ds.bind())
         throw new RuntimeException("directory bind failed");

      String appSrvPath = Utils.findDirectoryNodePath(ds,
                                                      "appservers/" + appServer,
                                                      "appserver",
                                                      false);

      if (appSrvPath == null)
      {
         throw new IllegalArgumentException(
            "No appserver named '" + appServer +"' is configured in the directory!");
      }
      
      Boolean auto = ds.getNodeBoolean(appSrvPath, "auto_start");
      String mode = ds.getNodeString(appSrvPath, "operating_mode");

      this.autoStart       = (auto != null && auto);
      this.userDir         = ds.getNodeString(appSrvPath, "user_dir");
      this.jvmArguments    = ds.getNodeStrings(appSrvPath, "jvm_args");
      this.operatingMode   = AppServerOperatingMode.fromString(mode);
      this.requestTimeout  = ds.getNodeInteger(appSrvPath, "request_timeout");
      this.autoTrimTimeout = ds.getNodeInteger(appSrvPath, "auto_trim_timeout");

      this.propath       = ds.getNodeString(appSrvPath, "propath");
      this.initialAgents = ds.getNodeInteger(appSrvPath, "initial_agents");
      this.minAgents     = ds.getNodeInteger(appSrvPath, "min_agents");
      this.maxAgents     = ds.getNodeInteger(appSrvPath, "max_agents");
      this.activate      = ds.getNodeString(appSrvPath, "activate");
      this.deactivate    = ds.getNodeString(appSrvPath, "deactivate");
      this.connect       = ds.getNodeString(appSrvPath, "connect");
      this.disconnect    = ds.getNodeString(appSrvPath, "disconnect");
      if (operatingMode.isSessionFree() && LOG.isWarnEnabled())
      {
         if (LOG.isWarnEnabled() && connect != null && connect.length() > 0)
         {
            LOG.warn("Connect procedure specified for appserver " + appServer + 
                      " in Session-free operating mode");
         }
         if (LOG.isWarnEnabled() && disconnect != null && disconnect.length() > 0)
         {
            LOG.warn("Disconnect procedure specified for appserver " + appServer + 
                     " in Session-free operating mode");
         }
      }
      this.startup          = ds.getNodeString(appSrvPath, "startup");
      this.shutdown         = ds.getNodeString(appSrvPath, "shutdown");
      this.startupParameter = ds.getNodeString(appSrvPath, "startup_parameter");

      // last use of the directory, end our session with it
      ds.unbind();
   }
   
   /**
    * Getter for the {@link #autoStart} field.
    * 
    * @return   See above.
    */
   public synchronized boolean isAutoStart()
   {
      return autoStart;
   }

   /**
    * Setter for the {@link #autoStart} field.
    * 
    * @param    autoStart
    *           The new value for the {@link #autoStart} field.
    */
   public synchronized void setAutoStart(boolean autoStart)
   {
      this.autoStart = autoStart;
   }

   /**
    * Getter for the {@link #userDir} field.
    * 
    * @return   See above.
    */
   public synchronized String getUserDir()
   {
      return userDir;
   }

   /**
    * Setter for the {@link #userDir} field.
    * 
    * @param    userDir
    *           The new value for the {@link #userDir} field.
    */
   public synchronized void setUserDir(String userDir)
   {
      this.userDir = userDir;
   }

   /**
    * Getter for the {@link #jvmArguments} field.
    * 
    * @return   See above.
    */
   public synchronized String[] getJvmArguments()
   {
      return jvmArguments == null ? new String[0] 
                                  : Arrays.copyOf(jvmArguments, jvmArguments.length);
   }

   /**
    * Setter for the {@link #jvmArguments} field.
    * 
    * @param    jvmArguments
    *           The new value for the {@link #jvmArguments} field.
    */
   public synchronized void setJvmArguments(String[] jvmArguments)
   {
      this.jvmArguments = Arrays.copyOf(jvmArguments, jvmArguments.length);
   }

   /**
    * Getter for the {@link #appServer} field.
    * 
    * @return   See above.
    */
   public synchronized String getAppServer()
   {
      return appServer;
   }

   /**
    * Setter for the {@link #appServer} field.
    * 
    * @param    appServer
    *           The new value for the {@link #appServer} field.
    */
   public synchronized void setAppServer(String appServer)
   {
      this.appServer = appServer;
   }

   /**
    * Getter for the {@link #operatingMode} field.
    * 
    * @return   See above.
    */
   public synchronized AppServerOperatingMode getOperatingMode()
   {
      return operatingMode;
   }

   /**
    * Setter for the {@link #operatingMode} field.
    * 
    * @param    operatingMode
    *           The new value for the {@link #operatingMode} field.
    */
   public synchronized void setOperatingMode(AppServerOperatingMode operatingMode)
   {
      this.operatingMode = operatingMode;
   }

   /**
    * Getter for the {@link #requestTimeout} field.
    * 
    * @return   See above.
    */
   public synchronized Integer getRequestTimeout()
   {
      return requestTimeout;
   }

  /**
   * Setter for the {@link #requestTimeout} field.
   * 
   * @param    requestTimeout
   *           The new value for the {@link #requestTimeout} field.
   */
   public synchronized void setRequestTimeout(Integer requestTimeout)
   {
      this.requestTimeout = requestTimeout;
   }

   /**
    * Getter for the {@link #autoTrimTimeout} field.
    * 
    * @return   See above.
    */
   public synchronized Integer getAutoTrimTimeout()
   {
      return autoTrimTimeout;
   }

   /**
    * Setter for the {@link #autoTrimTimeout} field.
    * 
    * @param    autoTrimTimeout
    *           The new value for the {@link #autoTrimTimeout} field.
    */
   public synchronized void setAutoTrimTimeout(Integer autoTrimTimeout)
   {
      this.autoTrimTimeout = autoTrimTimeout;
   }

   /**
    * Getter for the {@link #propath} field.
    * 
    * @return   See above.
    */
   public synchronized String getPropath()
   {
      return propath;
   }

   /**
    * Setter for the {@link #propath} field.
    * 
    * @param    propath
    *           The new value for the {@link #propath} field.
    */
   public synchronized void setPropath(String propath)
   {
      this.propath = propath;
   }

   /**
    * Getter for the {@link #initialAgents} field.
    * 
    * @return   See above.
    */
   public synchronized Integer getInitialAgents()
   {
      return initialAgents;
   }

   /**
    * Setter for the {@link #initialAgents} field.
    * 
    * @param    initialAgents
    *           The new value for the {@link #initialAgents} field.
    */
   public synchronized void setInitialAgents(Integer initialAgents)
   {
      this.initialAgents = initialAgents;
   }

   /**
    * Getter for the {@link #minAgents} field.
    * 
    * @return   See above.
    */
   public synchronized Integer getMinAgents()
   {
      return minAgents;
   }

   /**
    * Setter for the {@link #minAgents} field.
    * 
    * @param    minAgents
    *           The new value for the {@link #minAgents} field.
    */
   public synchronized void setMinAgents(Integer minAgents)
   {
      this.minAgents = minAgents;
   }

   /**
    * Getter for the {@link #maxAgents} field.
    * 
    * @return   See above.
    */
   public synchronized Integer getMaxAgents()
   {
      return maxAgents;
   }

   /**
    * Setter for the {@link #maxAgents} field.
    * 
    * @param    maxAgents
    *           The new value for the {@link #maxAgents} field.
    */
   public synchronized void setMaxAgents(Integer maxAgents)
   {
      this.maxAgents = maxAgents;
   }

   /**
    * Getter for the {@link #activate} field.
    * 
    * @return   See above.
    */
   public synchronized String getActivate()
   {
      return activate;
   }

   /**
    * Setter for the {@link #activate} field.
    * 
    * @param    activate
    *           The new value for the {@link #activate} field.
    */
   public synchronized void setActivate(String activate)
   {
      this.activate = activate;
   }

   /**
    * Getter for the {@link #deactivate} field.
    * 
    * @return   See above.
    */
   public synchronized String getDeactivate()
   {
      return deactivate;
   }

   /**
    * Setter for the {@link #deactivate} field.
    * 
    * @param    deactivate
    *           The new value for the {@link #deactivate} field.
    */
   public synchronized void setDeactivate(String deactivate)
   {
      this.deactivate = deactivate;
   }

   /**
    * Getter for the {@link #connect} field.
    * 
    * @return   See above.
    */
   public synchronized String getConnect()
   {
      return connect;
   }

   /**
    * Setter for the {@link #connect} field.
    * 
    * @param    connect
    *           The new value for the {@link #connect} field.
    */
   public synchronized void setConnect(String connect)
   {
      this.connect = connect;
   }

   /**
    * Getter for the {@link #disconnect} field.
    * 
    * @return   See above.
    */
   public synchronized String getDisconnect()
   {
      return disconnect;
   }

   /**
    * Setter for the {@link #disconnect} field.
    * 
    * @param    disconnect
    *           The new value for the {@link #disconnect} field.
    */
   public synchronized void setDisconnect(String disconnect)
   {
      this.disconnect = disconnect;
   }

   /**
    * Getter for the {@link #startup} field.
    * 
    * @return   See above.
    */
   public synchronized String getStartup()
   {
      return startup;
   }

   /**
    * Setter for the {@link #startup} field.
    * 
    * @param    startup
    *           The new value for the {@link #startup} field.
    */
   public synchronized void setStartup(String startup)
   {
      this.startup = startup;
   }

   /**
    * Getter for the {@link #shutdown} field.
    * 
    * @return   See above.
    */
   public synchronized String getShutdown()
   {
      return shutdown;
   }

   /**
    * Setter for the {@link #shutdown} field.
    * 
    * @param    shutdown
    *           The new value for the {@link #shutdown} field.
    */
   public synchronized void setShutdown(String shutdown)
   {
      this.shutdown = shutdown;
   }

   /**
    * Getter for the {@link #startupParameter} field.
    * 
    * @return   See above.
    */
   public synchronized String getStartupParameter()
   {
      return startupParameter;
   }

   /**
    * Setter for the {@link #startupParameter} field.
    * 
    * @param    startupParameter
    *           The new value for the {@link #startupParameter} field.
    */
   public synchronized void setStartupParameter(String startupParameter)
   {
      this.startupParameter = startupParameter;
   }
}